package dev.fyloz.colorrecipesexplorer.databasemanager

import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.ShouldSpec
import io.kotest.datatest.withData
import io.kotest.matchers.shouldBe
import io.mockk.every
import io.mockk.mockk
import io.mockk.mockkStatic
import org.slf4j.Logger
import java.sql.Connection
import java.sql.DriverManager
import java.sql.SQLException

class DatabaseTest : ShouldSpec({
    val logger = mockk<Logger> {
        every { debug(any()) } answers {}
    }
    val properties = databaseUpdaterProperties()
    val context = databaseContext(properties, logger)

    fun connectionMock(url: String) = every {
        DriverManager.getConnection(url, properties.username, properties.password)
    }

    fun serverConnectionMock() = connectionMock(properties.url)
    fun databaseConnectionMock() = connectionMock(properties.url + properties.dbName)

    context("CreDatabase") {
        mockkStatic(DriverManager::class.qualifiedName!!) {
            fun withDatabase(test: CreDatabase.() -> Unit) {
                serverConnectionMock() returns mockk()
                CreDatabase(context).test()
            }

            context("::new") {
                should("throws Connection exception when the url is invalid") {
                    serverConnectionMock() throws SQLException()

                    val exception = shouldThrow<CreDatabaseException.Connection> { CreDatabase(context) }
                    exception.url shouldBe properties.url
                }

                should("initialize a connection to the database's server") {
                    val connection = mockk<Connection>()
                    serverConnectionMock() returns connection

                    val database = CreDatabase(context)
                    database.serverConnection shouldBe connection
                }
            }

            context("::databaseConnection") {
                val databaseConnectionUrl = properties.url + properties.dbName

                should("throws Connection exception when the url is invalid") {
                    withDatabase {
                        databaseConnectionMock() throws SQLException()

                        val exception = shouldThrow<CreDatabaseException.Connection> { databaseConnection }
                        exception.url shouldBe databaseConnectionUrl
                    }
                }

                should("initialize a connection to the database") {
                    withDatabase {
                        val connection = mockk<Connection>()
                        databaseConnectionMock() returns connection

                        databaseConnection shouldBe connection
                    }
                }
            }

            context("::exists should be") {
                fun withDatabaseCatalog(catalog: String?, test: CreDatabase.() -> Unit) {
                    withDatabase {
                        every { serverConnection.metaData } returns mockk {
                            every { catalogs } returns mockk {
                                every { next() } answers { catalog != null } andThen false
                                every { getString(1) } answers { catalog } andThen null
                            }
                        }
                        test()
                    }
                }

                should("be true when database's catalogs contains database's name") {
                    withDatabaseCatalog(properties.dbName) {
                        exists shouldBe true
                    }
                }

                should("be false when database's catalogs does not contain given database's name") {
                    withDatabaseCatalog("anotherName") {
                        exists shouldBe false
                    }
                }

                should("be false when database's catalogs are empty") {
                    withDatabaseCatalog(null) {
                        exists shouldBe false
                    }
                }
            }

            context("::metadataTableExists should be") {
                withData(
                    mapOf(
                        "true when query returns a table" to true,
                        "false when query returns nothing" to false
                    )
                ) { exists ->
                    withDatabase {
                        every { serverConnection.metaData } returns mockk {
                            every {
                                getTables(properties.dbName, null, METADATA_TABLE_NAME, arrayOf("TABLE"))
                            } returns mockk {
                                every { next() } answers { exists } andThen false
                            }
                        }

                        metadataTableExists shouldBe exists
                    }
                }
            }

            context("::version") {
                fun withMockedDatabase(
                    databaseExists: Boolean,
                    metadataTableExists: Boolean,
                    versionRowExists: Boolean,
                    test: CreDatabase.() -> Unit
                ) {
                    withDatabase {
                        // Mocking lazy properties does not seems to work, so we need to mock everything
                        every { serverConnection.metaData } returns mockk {
                            every { catalogs } returns mockk {
                                every { next() } answers { databaseExists } andThen false
                                every { getString(1) } answers { properties.dbName }
                            }
                            every {
                                getTables(properties.dbName, null, METADATA_TABLE_NAME, arrayOf("TABLE"))
                            } returns mockk {
                                every { next() } returns metadataTableExists
                            }
                        }
                        databaseConnectionMock() returns mockk {
                            every { createStatement() } returns mockk {
                                every { executeQuery(QUERY_SELECT_VERSION) } returns mockk {
                                    every { next() } answers { versionRowExists } andThen false
                                    every { getInt(COLUMN_VALUE_NAME) } returns properties.targetVersion
                                }
                            }
                        }
                        test()
                    }
                }

                should("equals database's version") {
                    withMockedDatabase(databaseExists = true, metadataTableExists = true, versionRowExists = true) {
                        version shouldBe properties.targetVersion
                    }
                }

                should("throw UnknownDatabase exception when database doesn't exists") {
                    withMockedDatabase(databaseExists = false, metadataTableExists = true, versionRowExists = true) {
                        val exception = shouldThrow<CreDatabaseException.UnknownDatabase> { version }
                        exception.name shouldBe properties.dbName
                    }
                }

                should("be 0 when the metadata table does not exist") {
                    withMockedDatabase(databaseExists = true, metadataTableExists = false, versionRowExists = true) {
                        version shouldBe 0
                    }
                }

                should("be 0 when the version row does not exist") {
                    withMockedDatabase(databaseExists = true, metadataTableExists = true, versionRowExists = false) {
                        version shouldBe 0
                    }
                }
            }
        }
    }
})
