package dev.fyloz.colorrecipesexplorer.databasemanager

import io.kotest.assertions.throwables.shouldThrow
import io.kotest.core.spec.style.ShouldSpec
import io.kotest.datatest.withData
import io.kotest.matchers.shouldBe
import io.mockk.clearMocks
import io.mockk.every
import io.mockk.mockk
import java.io.File
import java.util.*

class PropertiesTest : ShouldSpec({
    val targetVersion = 0
    val url = "jdbc:driver://url:3306/"
    val dbName = "database"
    val username = "user"
    val password = "pass"

    context("database's url regular expression") {
        context("matches valid urls") {
            withData(
                "jdbc:mysql://host:9999/",
                "jdbc:h2://host:9999/",
                "jdbc:mysql://127.0.0.1:9999/",
                "jdbc:mysql://host:9999/",
                "jdbc:mysql://host/"
            ) { url ->
                url.matches(DATABASE_URL_REGEX) shouldBe true
            }
        }

        context("does not match invalid urls") {
            withData(
                "mysql://host:9999/",
                "jdbc://host:9999/",
                "jdbc:mysql:host:9999/",
                "jdbc:mysql/",
                "host:9999",
                "jdbc:mysql://host:9999/database",
                "jdbc:mysql://host"
            ) { url ->
                url.matches(DATABASE_URL_REGEX) shouldBe false
            }
        }

        should("not matches empty url") {
            "".matches(DATABASE_URL_REGEX) shouldBe false
        }
    }

    context("DatabaseUpdaterProperties::new throws") {
        context("InvalidTargetVersion exception with invalid database target versions") {
            withData(
                mapOf(
                    "negative" to -1,
                    "too large" to LATEST_DATABASE_VERSION + 1
                )
            ) { targetVersion ->
                val exception =
                    shouldThrow<InvalidDatabaseUpdaterPropertiesException.InvalidTargetVersion> {
                        DatabaseUpdaterProperties(targetVersion, url, dbName, username, password)
                    }
                exception.targetVersion shouldBe targetVersion
            }
        }

        context("InvalidUrl exception with invalid database urls") {
            withData(
                "mysql://host:9999/",
                "jdbc:mysql://:9999"
            ) { url ->
                val exception =
                    shouldThrow<InvalidDatabaseUpdaterPropertiesException.InvalidUrl> {
                        DatabaseUpdaterProperties(targetVersion, url, dbName, username, password)
                    }
                exception.url shouldBe url
            }
        }
    }

    context("DatabaseUpdaterProperties DSL from properties") {
        val properties = mockk<Properties>()
        every { properties.getProperty(PROPERTY_URL) } returns url
        every { properties.getProperty(PROPERTY_DB_NAME) } returns dbName
        every { properties.getProperty(PROPERTY_USERNAME) } returns username
        every { properties.getProperty(PROPERTY_PASSWORD) } returns password

        fun withTargetVersion(version: String?, test: () -> Unit) {
            every { properties.getProperty(PROPERTY_TARGET_VERSION) } returns version
            test()
        }

        fun withTargetVersion(version: Int, test: () -> Unit) = withTargetVersion(version.toString(), test)

        should("throw TargetVersionFormat with non-numeric target version") {
            val invalidTargetVersion = "zero"
            withTargetVersion(invalidTargetVersion) {
                val exception = shouldThrow<InvalidDatabaseUpdaterPropertiesException.TargetVersionFormat> {
                    databaseUpdaterProperties(properties)
                }
                exception.value shouldBe invalidTargetVersion
            }
        }

        should("return instance with correct values") {
            withTargetVersion(targetVersion) {
                with(databaseUpdaterProperties(properties)) {
                    this.targetVersion shouldBe targetVersion
                    this.url shouldBe url
                    this.dbName shouldBe dbName
                    this.username shouldBe username
                    this.password shouldBe password
                }
            }
        }

        should("throw MissingProperty when a property is not defined") {
            withTargetVersion(null) {
                val exception = shouldThrow<InvalidDatabaseUpdaterPropertiesException.MissingProperty> {
                    databaseUpdaterProperties(properties)
                }
                exception.property shouldBe PROPERTY_TARGET_VERSION
            }
        }

        afterEach {
            clearMocks(properties)
        }
    }

    context("DatabaseUpdaterProperties DSL from file") {
        val file = mockk<File>()
        val fileAbsolutePath = "/config/du.properties"
        every { file.absolutePath } returns fileAbsolutePath

        should("throw PropertiesNotFound when the file does not exists") {
            every { file.exists() } returns false
            every { file.isFile } returns true
            val exception = shouldThrow<InvalidDatabaseUpdaterPropertiesException.PropertiesNotFound> {
                databaseUpdaterProperties(file)
            }
            exception.path shouldBe fileAbsolutePath
        }

        should("throw PropertiesNotFound when the file is not a file") {
            every { file.exists() } returns true
            every { file.isFile } returns false
            val exception = shouldThrow<InvalidDatabaseUpdaterPropertiesException.PropertiesNotFound> {
                databaseUpdaterProperties(file)
            }
            exception.path shouldBe fileAbsolutePath
        }
    }
})
